#ifndef __LIBLFDS_H

  /***** abstraction *****/

  /***** defines *****/
  #if (defined _WIN64 && defined _MSC_VER)
    // TRD : 64-bit Windows with the Microsoft C compiler, any CPU
    #include <assert.h>
    #include <stdio.h>
    #include <stdlib.h>
    #include <windows.h>
    typedef unsigned __int64      atom_t;
    #define INLINE                extern __forceinline
    #define ALIGN(alignment)      __declspec( align(alignment) )
    #define ALIGN_SINGLE_POINTER  8
    #define ALIGN_DOUBLE_POINTER  16
  #endif

  #if (!defined _WIN64 && defined _WIN32 && defined _MSC_VER)
    // TRD : 32-bit Windows with the Microsoft C compiler, any CPU
    #include <assert.h>
    #include <stdio.h>
    #include <stdlib.h>
    #include <windows.h>
    typedef unsigned long int     atom_t;
    #define INLINE                extern __forceinline
    #define ALIGN(alignment)      __declspec( align(alignment) )
    #define ALIGN_SINGLE_POINTER  4
    #define ALIGN_DOUBLE_POINTER  8
  #endif

  #if (defined __unix__ && defined __x86_64__ && __GNUC__)
    // TRD : any UNIX with GCC on x64
    #define _XOPEN_SOURCE 600
    #include <assert.h>
    #include <stdio.h>
    #include <stdlib.h>
    typedef unsigned long long int   atom_t;
    #define INLINE                   inline
    #define ALIGN(alignment)         __attribute__( (aligned(alignment)) )
    #define ALIGN_SINGLE_POINTER     8
    #define ALIGN_DOUBLE_POINTER     16
  #endif

  #if (defined __unix__ && defined __i686__ && __GNUC__)
    // TRD : any UNIX with GCC on x86
    #define _XOPEN_SOURCE 600
    #include <assert.h>
    #include <stdio.h>
    #include <stdlib.h>
    typedef unsigned long int     atom_t;
    #define INLINE                inline
    #define ALIGN(alignment)      __attribute__( (aligned(alignment)) )
    #define ALIGN_SINGLE_POINTER  4
    #define ALIGN_DOUBLE_POINTER  8
  #endif

  /***** public prototypes *****/
  void abstraction_aligned_free( void *memory );
  void *abstraction_aligned_malloc( size_t size, size_t align_in_bytes );
  INLINE unsigned char abstraction_dcas( volatile atom_t *destination, atom_t *exchange, atom_t *compare );
  INLINE atom_t abstraction_increment( atom_t *value );





  /***** freelist *****/

  /***** enums *****/
  enum freelist_query_type
  {
    FREELIST_QUERY_ELEMENT_COUNT
  };

  /***** public prototypes *****/
  atom_t freelist_new( void **freelist_state, atom_t number_elements, int (*user_data_init_function)(void **user_data, void *user_state), void *user_state );
  void freelist_delete( void *freelist_state, void (*user_data_delete_function)(void *user_data, void *user_state), void *user_state );

  atom_t freelist_new_elements( void *freelist_state, atom_t number_elements );

  void *freelist_pop( void *freelist_state, void **freelist_element );
  void *freelist_guaranteed_pop( void *freelist_state, void **freelist_element );
  void freelist_push( void *freelist_state, void *freelist_element );

  void *freelist_get_user_data_from_element( void *freelist_element, void **user_data );
  void freelist_set_user_data_in_element( void *freelist_element, void *user_data );

  void freelist_query( void *freelist_state, enum freelist_query_type query_type, void *query_data );





  /***** queue *****/

  /***** defines *****/
  enum queue_query_type
  {
    QUEUE_QUERY_ELEMENT_COUNT
  };

  /***** public prototypes *****/
  atom_t queue_new( void **queue_state, atom_t number_elements );
  void queue_delete( void *queue_state, void (*user_data_delete_function)(void *user_data, void *user_state), void *user_state );

  int queue_enqueue( void *queue_state, void *user_data );
  int queue_guaranteed_enqueue( void *queue_state, void *user_data );
  int queue_dequeue( void *queue_state, void **user_data );

  void queue_query( void *queue_state, enum queue_query_type query_type, void *query_result );





  /***** ringbuffer *****/

  /***** public prototypes *****/
  atom_t ringbuffer_new( void **ringbuffer_state, atom_t number_elements, int (*user_data_init_function)(void **user_data, void *user_state), void *user_state );
  void ringbuffer_delete( void *ringbuffer_state, void (*user_data_delete_function)(void *user_data, void *user_state), void *user_state );

  void *ringbuffer_get_current_read_element( void *ringbuffer_state, void **ringbuffer_element );
  void ringbuffer_get_current_write_element( void *ringbuffer_state, void **ringbuffer_element );

  void ringbuffer_put_read_element( void *ringbuffer_state, void *ringbuffer_element );
  void ringbuffer_put_write_element( void *ringbuffer_state, void *ringbuffer_element );

  void *ringbuffer_get_user_data_from_element( void *ringbuffer_element, void **user_data );
  void ringbuffer_set_user_data_in_element( void *ringbuffer_element, void *user_data );





  /***** stack *****/

  /***** enums *****/
  enum stack_query_type
  {
    STACK_QUERY_ELEMENT_COUNT
  };

  /***** public prototypes *****/
  atom_t stack_new( void **stack_state, atom_t number_elements );
  void stack_delete( void *stack_state, void (*user_data_delete_function)(void *user_data, void *user_state), void *user_state );

  void stack_clear( void *stack_state, void (*user_data_clear_function)(void *user_data, void *user_state), void *user_state );

  int stack_push( void *stack_state, void *user_data );
  int stack_guaranteed_push( void *stack_state, void *user_data );
  int stack_pop( void *stack_state, void **user_data );

  void stack_query( void *stack_state, enum stack_query_type query_type, void *query_result );

  void stack_get_user_data_from_element( void *stack_element, void **user_data );





  #define __LIBLFDS_H

#endif

